#!/usr/bin/env python3
from whalelinter.app          import App
from whalelinter.dispatcher   import Dispatcher
from whalelinter.token        import Maintainer
from whalelinter.commands.apt import Apt

import whalelinter.commands.common


class Checker(object):
    def __init__(self, lst):
        self.lines              = lst
        self.line_number        = 0
        self.consecutive_run    = 0

    def check(self):
        dispatcher = Dispatcher()

        for self.line_number, line in self.lines:
            dispatcher.react(line, self.line_number)
            self.must_be_present(line[0])
            self.must_be_unique(line[0])
            self.should_be_present(line[0])

        if Apt._has_been_used:
            App._collecter.throw(3000, keys={'line': Apt._has_been_used})

        for mandatory_token in App._mandatory_tokens:
            App._collecter.throw(1000)

        for recommended_token in App._recommended_tokens:
            App._collecter.throw(3001, keys={'token': recommended_token})

    def must_be_present(self, token):
        if token in App._mandatory_tokens:
            App._mandatory_tokens.remove(token)

    def should_be_present(self, token):
        if token in App._recommended_tokens:
            App._recommended_tokens.remove(token)

    def must_be_unique(self, token):
        if token in App._unique_tokens:
            App._unique_tokens[token] += 1
            if App._unique_tokens[token] > 1:
                App._collecter.throw(1001, self.line_number, keys={'token': token})



#!/usr/bin/env python3

from whalelinter.app import App
from whalelinter.dispatcher import Dispatcher
from whalelinter.commands.command   import Command
# from whalelinter.commands.apt import Apt


class Token:
    def __init__(self, name, payload, line):
        self.name    = name
        self.payload = payload
        self.line    = line

        self.pointless_commands = App._pointless_commands


@Dispatcher.register(token='add')
class Add(Token):
    def __init__(self, payload, line):
        Token.__init__(self, __class__, payload, line)
        self.is_present()
        self.download_from_url()

    def is_present(self):
        App._collecter.throw(2006, self.line)
        return True

    def download_from_url(self):
        if ('http://' in self.payload[0] or 'https://' in self.payload[0]):
            App._collecter.throw(3004, self.line)
            return True
        return False


@Dispatcher.register(token='expose')
class Expose(Token):
    def __init__(self, ports, line):
        Token.__init__(self, __class__, ports, line)
        for port in ports:
            self.is_in_range(port)
            self.is_tcp_or_udp(port)

    def is_in_range(self, port):
        if '/' in port:
            port = port.split('/')[0]

        if int(port) < 1 or int(port) > 65535:
            App._collecter.throw(2005, self.line, keys={'port': port})
            return False
        return True

    def is_tcp_or_udp(self, port):
        if '/' in port:
            if port.split('/')[1] != 'tcp' and port.split('/')[1] != 'udp':
                App._collecter.throw(2009, self.line)
                return False
        return True


@Dispatcher.register(token='maintainer')
class Maintainer(Token):
    def __init__(self, payload, line):
        Token.__init__(self, __class__, payload, line)


@Dispatcher.register(token='run')
class Run(Token):
    def __init__(self, payload, line):
        Token.__init__(self, __class__, payload, line)
        self.is_pointless()

    def is_pointless(self):
        if self.payload[0] in self.pointless_commands:
            App._collecter.throw(2003, self.line, keys={'command': self.payload[0]})
            return True
        return False


@Dispatcher.register(token='from')
class SourceImage(Token):
    def __init__(self, payload, line):
        Token.__init__(self, __class__, payload, line)
        self.is_too_long()
        self.has_no_tag()
        self.has_latest_tag()

    def is_too_long(self):
        if len(self.payload) > 1:
            App._collecter.throw(1002, self.line, keys={'command': self.payload})
            return True
        return False

    def has_no_tag(self):
        if ':' not in self.payload[0]:
            App._collecter.throw(2000, self.line, keys={'image': self.payload[0]})
            return True
        return False

    def has_latest_tag(self):
        if ':latest' in self.payload[0]:
            App._collecter.throw(2001, self.line)
            return True
        return False


@Dispatcher.register(token='user')
class User(Token):
    def __init__(self, payload, line):
        Token.__init__(self, __class__, payload, line)

    def is_becoming_root(self):
        if self.payload[0] == 'root':
            App._collecter.throw(2007, self.line)
            return True
        return False


@Dispatcher.register(token='workdir')
class Workdir(Token):
    def __init__(self, payload, line):
        Token.__init__(self, __class__, payload, line)

    def has_relative_path(self):
        if not payload[0].startswith('/'):
            App._collecter.throw(2004, self.line)
            return True
        return False



#!/usr/bin/env python3
from whalelinter.app              import App
from whalelinter.dispatcher       import Dispatcher
from whalelinter.commands.command import PackageManager


@Dispatcher.register(token='run', command='apt-get')
@Dispatcher.register(token='run', command='apt')
class Apt(PackageManager):

    _callbacks      = {}
    _has_been_used  = 0

    def __init__(self, **kwargs):
        PackageManager.__init__(self, kwargs.get('token'), kwargs.get('command'), kwargs.get('args'), kwargs.get('lineno'))

        Apt.register(self)(type(self).install)
        Apt.register(self)(type(self).is_parameter_present, name='install', parameter='-y', args=kwargs.get('args'))
        Apt.register(self)(type(self).is_parameter_present, name='install', parameter='--no-install-recommends', args=kwargs.get('args'))
        Apt.register(self)(type(self).upgrade)
        Apt.register(self)(type(self).dist_upgrade, name='dist-upgrade')

        Apt._has_been_used = self.lineno

        for method in self.methods:
            if self.subcommand == method:
                self.react(method)

    def install(self):
        for idx, package in enumerate(self.packages):
            if '=' not in package:
                App._collecter.throw(3003, self.lineno, keys={'package': package})
            else:
                self.packages[idx] = package.split('=')[0]

        if sorted(self.packages) != self.packages:
            App._collecter.throw(3002, self.lineno)

    def upgrade(self):
        App._collecter.throw(2008, self.lineno)

        for idx, package in enumerate(self.packages):
            if '=' not in package:
                App._collecter.throw(3003, self.lineno, keys={'package': package})
            else:
                packages[idx] = package.split('=')[0]

        if self.packages and sorted(self.packages) == self.packages:
            App._collecter.throw(3002, self.lineno)

    def dist_upgrade(self):
        App._collecter.throw(2011, self.lineno)



#!/usr/bin/env python3

import inspect
from whalelinter.app import App


class Command:
    _callbacks = {}

    @classmethod
    def register(cls, instance):
        def decorate(func, **kwargs):

            if 'name' not in kwargs:
                name = func.__name__
            else:
                name = kwargs.get('name')

            if (name not in cls._callbacks):
                cls._callbacks[name] = []

            # if 'args' in kwargs and name in .get(args):

            cls._callbacks[name].append({
                'instance': instance,
                'function': func,
                'kwargs'  : kwargs
            })

        return decorate

    def react(self, name):
        if name in self._callbacks:
            for callback in self._callbacks[name]:
                if (self == callback.get('instance')):
                    if callback.get('kwargs'):
                        callback.get('function')(callback.get('instance'), **callback.get('kwargs'))
                    else:
                        callback.get('function')(callback.get('instance'))


class PackageManager(Command):

    def find_packages(self):
        lst = list(self.rest)
        for arg in self.args:
            lst.remove(arg)

        subcommand = lst[0]
        lst.remove(lst[0])

        return (subcommand, lst)

    def is_parameter_present(self, **kwargs):
        if 'parameter' in kwargs and 'args' in kwargs:
            if kwargs.get('parameter') not in kwargs.get('args'):
                App._collecter.throw(2010, self.lineno, keys={
                    'parameter' : kwargs.get('parameter'),
                    'command'   : self.full_command
                })
                return False

        return True

    def __init__(self, token, command, rest, line_number):
        """
            RUN apt-get install -y name

            RUN             : token
            apt-get         : command
            install -y name  : rest

            install : subcommand
            -y      : args
            name     : packages

        """
        self.token          = token
        self.command        = command
        self.rest           = rest
        self.full_command   = self.command + ' ' + ' '.join(self.rest)

        self.args           = [arg for arg in self.rest if (arg.startswith('-') or arg.startswith('--'))]
        self.lineno         = line_number
        self.methods        = []

        self.subcommand, self.packages = self.find_packages()

        for method in inspect.getmembers(self, predicate=inspect.ismethod):
            if not method[0].startswith('__') \
                and method[0] is not 'register' \
                and method[0] is not 'find_packages' \
                and method[0] is not 'is_parameter_present' \
                and method[0] is not 'react':

                self.methods.append(method[0])



#!/usr/bin/env python3
import shlex
import operator
import re
import urllib.request
import os
from whalelinter.app import App


class Parser(object):
    def __init__(self, filename):
        if self.is_url(filename) is not None:
            response = urllib.request.urlopen(filename)
            if self.is_content_type_plain_text(response):
                self.file = response.read().decode('utf-8')
            else:
                print('ERROR: file format not supported\n')
        elif os.path.isfile(filename):
            self.file = open(filename, encoding='utf-8').read()
        elif self.is_github_repo(filename):
            filename = 'https://raw.githubusercontent.com/' + filename + '/master/Dockerfile'
            self.file = urllib.request.urlopen(filename).read().decode('utf-8')
        else:
            print('ERROR: file format not supported\n')

        self.TOKENS = App._config.get('all')

    def is_github_repo(self, filename):
        regex = re.compile(r'^[-_.0-9a-z]+/[-_.0-9a-z]+$', re.IGNORECASE)

        return True if regex.match(filename) is not None else False

    def is_url(self, filename):
        regex = re.compile(
            r'^(?:http|ftp)s?://' # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
            r'localhost|' #localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
            r'(?::\d+)?' # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE)

        return regex.match(filename)

    def is_content_type_plain_text(self, response):
        content_type = response.getheader('Content-Type')
        regex        = re.compile(r'text/plain')

        return True if regex.search(content_type) is not None else False

    def shlex_to_dictionnary(self):
        self.lexer                  = shlex.shlex(instream=self.file, posix=True)
        self.lexer.quotes           = '"'
        self.lexer.commenters       = '#'
        self.lexer.whitespace_split = True

        accumulator = []
        commands    = {}
        line        = 0

        for word in self.lexer:
            if word in self.TOKENS:
                if accumulator:
                    commands[line] = accumulator
                    accumulator = []

                line = self.lexer.lineno

            accumulator.append(word)

        commands[line] = accumulator

        return sorted(commands.items(), key=operator.itemgetter(0))



#!/usr/bin/env python3
from whalelinter.app import App

FUNCTION_CHAR = '_'


class Dispatcher:
    _callbacks = {}

    def __init__(self):
        self.consecutive_run = {
            'count': 0,
            'line' : 0
        }

    def process_chained_commands(self, command):
        idx, is_newline = self.find_first_command_chaining_symbol(command)

        rest    = command[idx + 1:]
        command = command[:idx]

        if rest[0] == '\n':
            del rest[0]

        if rest[0].startswith('\n'):
            rest[0] = rest[0][1:]

        return (command, rest, is_newline)

    def find_first_command_chaining_symbol(self, command):
        is_newline  = False
        logical_AND = next((x for x in command if x == '&&'), None)
        idx         = command.index(logical_AND) if logical_AND else 0

        if idx and (
            command[idx - 1] == '\n' or \
            command[idx + 1] == '\n' or \
            '\n' in command[idx - 1] or \
            '\n' in command[idx + 1]):

            is_newline = True

        logical_AND = next((x for x in command if x == '\n&&'), None)
        if logical_AND :
            if idx > 0:
                if command.index(logical_AND) < idx:
                    idx        = command.index(logical_AND)
                    is_newline = True
            else:
                idx        = command.index(logical_AND)
                is_newline = True

        return (idx, is_newline)


    def react(self, docker_command, lineno=None):
        """
        When a command is identified, find the corresponding callback and
        let it handle the command.

        self._callbacks = {
            'run' = {
                '_': func(self.caller, args), # represents the 'run' function
                'apt-get' = {
                    '_': func(self.caller, args),
                    'install' = {
                        '_': func(self.caller, args),
                    },
                    'upgrade' = {
                        '_': func(self.caller, args),
                    }
                }
            }
        }

        """
        rest = False
        token, args = docker_command[0].lower(), docker_command[1:]

        if '&&' in args or '\n&&' in args:
            args, rest, is_multiline_command = self.process_chained_commands(args)

        if token in self._callbacks:
            if (self._callbacks[token][FUNCTION_CHAR] is not None):
                self._callbacks[token][FUNCTION_CHAR](args, lineno)

            if token == 'run':
                self.consecutive_run['count'] += 1
                self.consecutive_run['line']   = lineno
            else:
                self.consecutive_run['count'] = 0

            command = args[0]
            args = args[1:]

            if command in self._callbacks[token]:

                if self._callbacks[token][command][FUNCTION_CHAR] is not None:
                    self._callbacks[token][command][FUNCTION_CHAR](token=token, command=command, args=args, lineno=lineno)

        if self.consecutive_run.get('count') > 1:
            App._collecter.throw(2012, self.consecutive_run.get('line'))

        if rest:
            if is_multiline_command:
                lineno += 1

            rest.insert(0, 'RUN')
            self.consecutive_run['count'] = 0
            self.react(rest, lineno)

    @classmethod
    def register(cls, func=None, token=None, command=None):
        """
        Decorator to register a new callback.

        A callback should have the same name as its corresponding command.
        It takes an instance of the calling Dispatcher object as its first
        argument and may accept an arbitrary number of positional arguments.
        If the decorated function contains a 'name' argument, the callback
        will be registered with this name instead of the function one.

        @Dispatcher.register(token='my_command', command='foo', subcommand='bar')
        def my_command_foo_bar(self, args):
        """

        if (not cls._callbacks):
            cls._callbacks = {x.lower(): {FUNCTION_CHAR: None, } for x in App._config.get('all')}

        if hasattr(func, '__call__'):
            token = func.__name__

        if token and command:
            if command not in cls._callbacks[token]:
                cls._callbacks[token][command] = {
                    FUNCTION_CHAR: None,
                }

        if token and command:
            def decorate(func):
                cls._callbacks[token][command][FUNCTION_CHAR] = func

                return func

        if token and not command:

            def decorate(func):
                cls._callbacks[token][FUNCTION_CHAR] = func

                return func

        return decorate



#!/usr/bin/env python3
import argparse
from pkg_resources import get_distribution, DistributionNotFound

from whalelinter.app       import App
from whalelinter.parser    import Parser
from whalelinter.checker   import Checker
from whalelinter.collecter import Collecter

__all__ = ['run']


def run():
    dist = get_distribution('whale-linter')
    __version__ = dist.version

    parser = argparse.ArgumentParser(description='A simple non professional Dockerfile linter')
    parser.add_argument('-i', '--ignore', default=None, help='Rule to ignore', action='append')
    parser.add_argument('-v', '--version', help='Print version', action='version', version='%(prog)s {}'.format(__version__))
    parser.add_argument('DOCKERFILE', help="The Dockerfile to lint (can be a file, a GitHub repo. or a direct URL)")
    args = parser.parse_args()

    print('\n{}\n'.format(args.DOCKERFILE))

    if args.ignore is None:
        args.ignore = []

    App._collecter = Collecter(App._config.get('rules'), args.ignore)

    parser  = Parser(args.DOCKERFILE)
    checker = Checker(parser.shlex_to_dictionnary())
    checker.check()

    App._collecter.display()

if __name__ == '__main__':
    run()



#!/usr/bin/env python3
import os
import json


class App:
    @staticmethod
    def load_configuration():
        config_file = os.path.dirname(os.path.realpath(__file__)) + '/config.json'
        with open(config_file) as data_file:
            data = json.load(data_file)

        return data

    _config             = load_configuration.__func__()
    _collecter          = None
    _mandatory_tokens   = _config.get('mandatory')
    _recommended_tokens = _config.get('recommended')
    _pointless_commands = _config.get('pointless_commands')
    _unique_tokens      = {x: 0 for x in _config.get('unique')}



#!/usr/bin/env python3
COLORS = {
    'BLUE'  : '\033[94m',
    'GREEN' : '\033[92m',
    'YELLOW': '\033[93m',
    'RED'   : '\033[91m',
    'ENDC'  : '\033[0m',
    'BOLD'  : '\033[1m',
}


class Log:
    def __init__(self, rule, line, keys):
        self.id       = rule.get('id')
        self.category = rule.get('category')
        self.message  = rule.get('message').format(**keys)

        if line is None:
            line = 0

        self.line = line

    def display(self, color, category_right_padding, line_number_right_padding):
        category_right_padding = category_right_padding * ' '
        line_number_right_padding = line_number_right_padding * ' '

        if self.line:
            print('  {}{}: {}{}{} : {}{}'.format(line_number_right_padding, self.line, COLORS.get(color.upper()), self.category, category_right_padding, COLORS.get('ENDC'), self.message))
        else:
            print('      {}{}{} : {}{}'.format(COLORS.get(color.upper()), self.category, category_right_padding, COLORS.get('ENDC'), self.message))


class Collecter:
    def __init__(self, rules, ignore=None):
        self.ignore      = ignore
        self.rules       = rules
        self.logs        = []
        self.log_classes = [
            {
                'level'      : 'critical',
                'color'      : 'red',
                'categories' : ['NotFound', 'TooMuch', 'TooLong', 'BadValue']
            },
            {
                'level'      : 'warning',
                'color'      : 'yellow',
                'categories' : ['BadPractice', 'Pointless']
            },
            {
                'level'      : 'enhancement',
                'color'      : 'blue',
                'categories' : ['BestPractice', 'Immutability', 'Maintainability']
            },
        ]

    def throw(self, id, line=None, keys={}):
        # rule = (item for item in self.rules if .get(id) == id)
        if str(id) not in self.ignore:
            for rule in self.rules:
                if rule.get('id') == str(id):
                    break

            log = Log(rule, line, keys)
            self.logs.append(log)

    def find_longest_category_name(self, level):
        length = 0
        for log_class in self.log_classes:
            if level == log_class.get('level'):
                for category in log_class.get('categories'):
                    if len(category) > length:
                        length = len(category)

        return length

    def find_highest_line_number(self):
        length = 0
        for log in self.logs:
            if len(str(log.line)) > length:
                length = len(str(log.line))

        return length

    def display(self):
        if self.logs:
            self.logs.sort(key=(lambda log: log.line))

            for log_class in self.log_classes:
                level      = log_class.get('level')
                color      = log_class.get('color').upper()
                categories = log_class.get('categories')

                print('{}{} :{}'.format(COLORS.get(color), level.upper(), COLORS.get('ENDC')))
                for log in self.logs:
                    if log.category in categories:
                        category_right_padding = self.find_longest_category_name(level) - len(log.category)
                        line_number_right_padding = self.find_highest_line_number() - len(str(log.line))
                        log.display(color, category_right_padding, line_number_right_padding)
                print()
        else:
            print('{}{}{}\n'.format(COLORS.get('GREEN'), 'Everything is good', COLORS.get('ENDC')))



#!/usr/bin/env python3
import re

from whalelinter.app                import App
from whalelinter.dispatcher         import Dispatcher
from whalelinter.commands.command   import Command
from whalelinter.commands.apt       import Apt


@Dispatcher.register(token='run', command='cd')
class Cd(Command):
    def __init__(self, **kwargs):
        App._collecter.throw(2002, self.line)
        return False


@Dispatcher.register(token='run', command='rm')
class Rm(Command):
    def __init__(self, **kwargs):
        rf_flags_regex    = re.compile("(-.*[rRf].+-?[rRf]|-[rR]f|-f[rR])")
        rf_flags          = True if [i for i in kwargs.get('args') if rf_flags_regex.search(i)] else False
        cache_path_regex  = re.compile("/var/lib/apt/lists(\/\*?)?")
        cache_path        = True if [i for i in kwargs.get('args') if cache_path_regex.search(i)] else False

        if rf_flags and cache_path:
            if (int(Apt._has_been_used) < int(kwargs.get('lineno'))):
                Apt._has_been_used = 0



#!/usr/bin/env python3
from whalelinter.app              import App
from whalelinter.dispatcher       import Dispatcher
from whalelinter.commands.command import PackageManager


@Dispatcher.register(token='run', command='apk')
class Apk(PackageManager):

    _callbacks      = {}

    def __init__(self, **kwargs):
        PackageManager.__init__(self, kwargs.get('token'), kwargs.get('command'), kwargs.get('args'), kwargs.get('lineno'))

        Apk.register(self)(type(self).add)

        for method in self.methods:
            if self.subcommand == method:
                self.react(method)

    def add(self):
        print(self.full_command)



#!/usr/bin/env python3

from setuptools import setup, find_packages
from os         import path
import sys

if sys.version_info < (3, 2):
    print('whale-linter requires at least Python 3.2 to run.')
    sys.exit(1)

here = path.abspath(path.dirname(__file__))

version = '0.0.4'

setup(
    name='whale-linter',
    version=version,
    packages=find_packages(),
    scripts=['bin/whale-linter'],
    author="Jerome Pin",
    author_email="jerome@jeromepin.fr",
    maintainer="Jerome Pin",
    maintainer_email="jerome@jeromepin.fr",
    description="A simple non professional Dockerfile linter",
    long_description='',
    include_package_data=True,
    license='MIT',
    url='https://github.com/jeromepin/whale-linter',
    download_url='https://github.com/jeromepin/whale-linter/tarball/' + version,
    keywords=['docker', 'linter'],
    platforms='Linux',
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        "Intended Audience :: System Administrators"
    ]
)
